#include <snort.h>

/*
* Snort parsing is a 3 step process.
*
* 1. Build a list of key/value pairs (for our matchers)
* 2. Optimise/fix that list of matchers, eg: make 'msg' data
*    become the argument for the target etc..
* 3. Actually build and add the rule(s), then cleanup the key/val
*    list.
*/

proc_print_out print_out;
proc_print_raw print_raw;

proc_add_rule		add_rule;
proc_add_criteria	add_criteria;
proc_set_target		set_target;
proc_add_parser		add_parser;
proc_rule_cleanup	rule_cleanup;

proc_list_append	list_append;
proc_list_remove	list_remove;

/* This global data isn't protected. that means
* this plugin is in NO WAY reentrant. */
char *target=NULL;
char *t_priv=NULL;
p_kv temp=NULL;
p_kv opts=NULL;

void snort_flip_rule(p_kv t)
{
	p_kv spt=NULL, dpt=NULL;
	p_kv src=NULL, dst=NULL;
	p_kv tmp;

	for ( tmp=temp; tmp; tmp=(p_kv)tmp->l.next ) {
		if ( !strcasecmp("SPT", tmp->key) ) {
			spt = tmp;
			continue;
		}
		if ( !strcasecmp("DPT", tmp->key) ) {
			dpt = tmp;
			continue;
		}
		if ( !strcasecmp("IP_SRC", tmp->key) ) {
			src = tmp;
			continue;
		}
		if ( !strcasecmp("IP_DST", tmp->key) ) {
			dst = tmp;
			continue;
		}
	}

	/* Actually swap round the values */
	if ( spt ) strcpy(spt->key, "DPT");
	if ( dpt ) strcpy(dpt->key, "SPT");
	if ( src ) strcpy(src->key, "IP_DST");
	if ( dst ) strcpy(dst->key, "IP_SRC");
}

int snort_add_rule(void)
{
	p_kv p, the_opts;
	p_rule r;
	
	char *real_t_priv;
	
	if ( !t_priv )
		real_t_priv="SNORT";
	else
		real_t_priv=t_priv;
	
	if ( !(r=malloc(sizeof(rule_t))) ) return 0;
	SAFE_INIT_RULE_PTR(r);

	/* Set the rules target */
	if ( !(set_target(r, target, real_t_priv)) ) {
		rule_cleanup(r);
		return 0;
	}
	
	for (p=temp; p; p=(p_kv)p->l.next) {
		if ( !strcasecmp("CONTENT", p->key) ) {
			the_opts=opts;
		}else{
			the_opts=NULL;
		}
		
		if ( !add_criteria(r, p->key, p->val, the_opts, p->flags) ) {
			rule_cleanup(r);
			return 0;
		}
	}
	
	if ( !add_rule(r) ) {
		rule_cleanup(r);
		return 0;
	}
	
	return 1;
}


void snort_kv_clean(p_kv k)
{
	p_kv foo=NULL, bar=NULL;

	for (foo=k; foo;) {
		if ( foo->key ) free(foo->key);
		if ( foo->val ) free(foo->val);
		bar=foo;
		foo=(p_kv)foo->l.next;
		free(bar);
	}
}

void snort_cleanup_internal(void)
{
	p_kv foo=NULL, bar=NULL;
	
	if ( target ) free(target);
	if ( t_priv ) free(t_priv);

	snort_kv_clean(temp);
	snort_kv_clean(opts);
	
	temp=NULL;
	opts=NULL;
}

int snort_link_element(const char *key, const char *val, int nval)
{
	p_kv n=NULL;
	
	if ( !key || !val ) return 0;

	if ( !(n=malloc(sizeof(kv_t))) ) return 0;

	if ( nval == NVAL_AUTO ) {
		if ( *val == '!' ) {
			val++;
			n->flags=1;
		}else{
			n->flags=0;
		}
	}else{
		n->flags = (nval) ? 1 : 0 ;
	}
	
	if ( !(n->key=strdup(key)) ) {
		free(n);
		return 0;
	}
	
	if ( !(n->val=strdup(val)) ) {
		free(n->key);
		free(n);
		return 0;
	}

	temp=list_append(n, temp);
	return 1;
}

int snort_optimise_rule(p_kv tokens)
{
	p_kv p,t;

	/* Turn the MSG param in to the private data */
	for (p=temp; p; ) {
		if ( !strcasecmp(p->key, "MSG" ) ) {
			if ( !p->l.prev ) temp=(p_kv)p->l.next;
			
			t=list_remove(p);
			p=(p_kv)p->l.next;
			t_priv=t->val;
			free(t->key);
			free(t);
			continue;
		}

		if ( !strcasecmp(p->key, "DEPTH") ||
			!strcasecmp(p->key, "OFFSET") ||
			!strcasecmp(p->key, "NOCASE") ) {

			if ( !p->l.prev ) temp=(p_kv)p->l.next;
			t=list_remove(p);

			opts=(p_kv)list_append(t, opts);
		}
		p=(p_kv)p->l.next;
	}
	
	return 1;
}

/* TODO:
* Allow the escaping of quote marks inside quoted strings
* Strip the quote marks off of quoted strings
* */
#define PARSE_CRAP	0
#define PARSE_KEY	1
#define PARSE_VAL	2
#define PARSE_END	3
int snort_parse_body(char *body)
{
	char *cur=NULL;
	int  cur_meaning=PARSE_CRAP;
	char *begin_key=NULL;
	char *begin_val=NULL;
	int keys=0;
	int quote=0;
	char *end;

	end=body+strlen(body);

	for ( cur=body; cur <= end; cur++ ) {
		switch ( cur_meaning ) {
			case PARSE_CRAP:
				if ( quote ) break;
				if ( *cur == '(' ) break;
				if ( *cur == ')' ) break;
				if ( !isblank(*cur) ) {
					cur_meaning=PARSE_KEY;
					begin_key=cur;
				}
				break;
			case PARSE_KEY:
				if ( quote ) break;
				if ( *cur == ':' ) {
					*cur='\0';
					cur_meaning=PARSE_VAL;
					begin_val=cur+1;
				}
				break;
			case PARSE_VAL:
				if ( *cur == '"' ) {
					if ( !quote ) begin_val=cur+1;

					if ( quote ) *cur='\0';
					
					quote=(quote) ? 0 : 1;
				}
				if ( quote ) break;
				if ( *cur == ';' || *cur==')') {
					*cur='\0';
					cur_meaning=PARSE_END;
					
				}
				break;
			case PARSE_END:
				snort_link_element(begin_key, begin_val, NVAL_NORMAL);
				keys++;
				cur_meaning=PARSE_CRAP;
				begin_val=NULL;
				begin_key=NULL;
				break;
		}
	}
	
	return keys ? 1 : 0;
}

snort_ret snort_parse_rule(char *rule)
{
	p_rule nrule;
	char *tokens[5];
	int i=0;
	char *work;
	snort_ret ret=SNORT_NULL;
	
	target=strdup(strtok(rule," \t"));
	if ( !target ) return SNORT_NULL;
	
	/* Get the head part tokenised:
	* target src spt -> dst dpt */
	while ( (tokens[i]=strtok(NULL," \t"))  && (i++<5) );
	if ( i != 6 ) return SNORT_NULL;
	
	/* Which way the rule goes, back, forth, bi */
	if ( !strcmp(tokens[3], "<-") ) { ret=SNORT_BACKWARDS; }
	if ( !strcmp(tokens[3], "->") ) { ret=SNORT_FORWARDS; }
	if ( !strcmp(tokens[3], "<>") ) { ret=SNORT_BIDIRECTIONAL; }
	
	/* Has to go back OR forth */
	if ( ret == SNORT_NULL ) return ret;
	
	/* Add the head stuff to the rule */
	snort_link_element("IP_PROTO", tokens[0], NVAL_NORMAL);

	if ( strcasecmp("any", tokens[1]) )
		snort_link_element("IP_SRC", tokens[1], NVAL_AUTO);
	if ( strcasecmp("any", tokens[2]) ) 
		snort_link_element("SPT", tokens[2], NVAL_AUTO);
	if ( strcasecmp("any", tokens[4]) ) 
		snort_link_element("IP_DST", tokens[4], NVAL_AUTO);
	if ( strcasecmp("any", tokens[5]) ) 
		snort_link_element("DPT", tokens[5], NVAL_AUTO);
	

	/* We now have the variable sized rule body
	* in this format:
	* (key:123; key2:"val"; key:"val |0d|";)  */
	
	if ( !(work=strdup(strtok(NULL, ""))) ) {
		return SNORT_NULL;
	}
	
	if ( !snort_parse_body(work) ) ret = SNORT_NULL;
	
	free(work);
	return ret;
}

int snort_loadfile(char *fn)
{
	FILE *f;
	char buf[4096];
	char *p=(char *)&buf;
	int rule=0, line=0;
	snort_ret ret;
	
	if ( !fn ) return 0;
	
	if ( (f=fopen(fn, "r")) == NULL ) {
		return 0;
	}

	print_out("Opening snort rulefile (%s)" CRLF, fn);

	while( fgets(p, 4096, f) ) {
		line++;
		if ( IS_RULE(buf) ) {
			rule++;
			ret = snort_parse_rule(p);
			
			if ( ret == SNORT_NULL ) {
				print_out("Parse error on rule %i (%s:%i)" CRLF,
					rule, fn, line);
				goto rule_dud;
			}

			if ( !snort_optimise_rule(temp) ) {
				print_out("Optimise error on rule %i (%s:%i)" CRLF,
					rule, fn, line);
				goto rule_dud;
			}

			if ( ret == SNORT_FORWARDS || ret == SNORT_BIDIRECTIONAL ) {
				if ( !snort_add_rule() ) {
					print_out("Rule error on rule %i (%s:%i)" CRLF,
						rule, fn, line);
					goto rule_dud;
				}
			}
			
			if ( ret == SNORT_BACKWARDS || ret == SNORT_BIDIRECTIONAL ) {
				snort_flip_rule(temp);
				if ( !snort_add_rule() ) {
					print_out("Rule error on rule %i (%s:%i)" CRLF,
						rule, fn, line);
					goto rule_dud;
				}
			}

rule_dud:
			/* Clean up our parsers little scratch area */
			snort_cleanup_internal();
		}
	}
	
	fclose(f);
	return 1;
}

uint _PLUGIN_INIT (p_plugin_in in, p_plugin_out out)
{
	PLUGIN_CHECK_PARAMS(in);

	/* Take the functions we need */
	print_out=in->print_out;
	print_raw=in->print_raw;
	add_rule=in->add_rule;
	add_criteria=in->add_criteria;
	set_target=in->set_target;
	add_parser=in->add_parser;
	rule_cleanup=in->rule_cleanup;
	list_append=in->list_append;
	list_remove=in->list_remove;
	
	/* Fill in a description about ourselves */
	out->name="ruleset/snort";
	out->desc="Loads snort ruleset files";
	out->ver_major=1;
	out->ver_minor=0;
	
	/* Tell firestorm what function to
	* call when it needs to load a 'Snort' file */
	add_parser("Snort", snort_loadfile);
	
	return PLUGIN_ERR_OK;
}

/* This is where we would cleanup our shit */
uint _PLUGIN_UNLOAD (void)
{
	return PLUGIN_ERR_OK;
}
